#!/usr/bin/env python3
# coding: utf-8
# Author: Arne Köhn <arne@chark.eu>,
#         minor changes by Christine Köhn <christine@chark.eu>
# License: MIT / Apache v2 dual licensed

from lxml import etree as ET
import os
import sys
import numpy as np


def extract_tokens(tier_elem, timestamps):
    return [(timestamps.index(x.get("end")), x.text)
            for x in tier_elem.findall("event")]


def read_xml_to_annotation(flo, level="ZH1"):
    root = ET.parse(flo)

    timestamps = [x.get("id") for x in root.findall("//common-timeline/tli")]

    origin_toks = extract_tokens(root.find("//tier[@display-name='ctok']"),
                                 timestamps)
    zh1_toks = extract_tokens(root.find("//tier[@display-name='"+level+"']"),
                              timestamps)

    sent = extract_tokens(root.find("//tier[@display-name='ctokS']"),
                          timestamps)

    sent_level = extract_tokens(root.find("//tier[@display-name='"+level+"S']"),
                                timestamps)

    sentence_idx = 0
    sentence_level_idx = 0
    zh1_idx = 0
    orig_idx = 0

    end_z = -1

    result = []
    sentence_lengths = []
    sentence_level_lengths = []

    curr_sentence_length = 0
    curr_sentence_level_length = 0
    
    while True:
        (end_o, word_o) = origin_toks[orig_idx]
        orig_idx += 1
        # consume orig tokens until we find a non-empty one
        while word_o is None and orig_idx < len(origin_toks):
            (end_o, word_o) = origin_toks[orig_idx]
            orig_idx += 1

        curr_sentence_length += 1
        if (sent[sentence_idx][0] <= end_o):
            sentence_lengths.append(curr_sentence_length)
            sentence_idx += 1
            curr_sentence_length = 0

        # at the end of orig; consume all corrections left
        if orig_idx == len(origin_toks):
            corr = "".join(x[1] for x in zh1_toks[zh1_idx:])
            curr_sentence_level_length += sum(x[1] != "" for x in zh1_toks[zh1_idx:])
            sentence_level_lengths.append(curr_sentence_level_length)
            result.append((word_o, corr))
            break

        # consume all preceeding correction tokens
        correction_string = ""
        while end_o > end_z:
            (end_z, word_z) = zh1_toks[zh1_idx]
            if word_z is not None and word_z != "":
                correction_string += word_z
                curr_sentence_level_length += 1
            if (sent_level[sentence_level_idx][0] <= end_o):
                sentence_level_lengths.append(curr_sentence_level_length)
                sentence_level_idx += 1
                curr_sentence_level_length = 0
            zh1_idx += 1

        result.append((word_o, correction_string))
    return {"changes": result,
            "sentence_lengths": sentence_lengths,
            "sentence_level_lengths": sentence_level_lengths}


def num_changes_between(tokens):
    return sum([0 if x[0] == x[1] else 1 for x in tokens])


def appendall(into, newdata):
    for k in newdata:
        if k not in into:
            into[k] = newdata[k]
        else:
            into[k] += newdata[k]


def compute_kappa(level="ZH1"):
    annotator_a = {}
    annotator_b = {}
    for fname in os.listdir(sys.argv[1] + "/set2/xml/annotator_a/"):
        aflo = open(sys.argv[1] + "/set2/xml/annotator_b/" + fname)
        bflo = open(sys.argv[1] + "/set2/xml/annotator_a/" + fname)
        appendall(annotator_b, read_xml_to_annotation(bflo, level))
        appendall(annotator_a, read_xml_to_annotation(aflo, level))

    a_changes = annotator_a["changes"]
    b_changes = annotator_b["changes"]

    numtokens = len(b_changes)
    changes_by_annotator_b = num_changes_between(b_changes)
    changes_by_annotator_a = num_changes_between(a_changes)

    changes_b_a = num_changes_between(
        [(x[0][1], x[1][1])
         for x in zip(b_changes, a_changes)])

    changes_b_a_positions = num_changes_between(
        [(x[0][1] == x[0][0], x[1][1] == x[1][0])
         for x in zip(b_changes, a_changes)])

    pr_a = (numtokens - changes_b_a) * 1.0 / numtokens
    pr_a_positions = (numtokens - changes_b_a_positions) * 1.0 / numtokens

    # probability of a change by a and b
    pr_ann_a = (changes_by_annotator_a * 1.0 / numtokens)
    pr_ann_b = (changes_by_annotator_b * 1.0 / numtokens)
    pr_e = (pr_ann_a * pr_ann_b) + (1 - pr_ann_a) * (1 - pr_ann_b)

    kappa = (pr_a - pr_e) / (1 - pr_e)
    kappa_positions = (pr_a_positions - pr_e) / (1 - pr_e)
    print("num tokens ctok", len(b_changes),
          "changes_by_annotator_b", changes_by_annotator_b,
          "changes_by_annotator_a", changes_by_annotator_a,
          "changes_b_a", changes_b_a,
          "changes_b_a_positions", changes_b_a_positions,
          "kappa", kappa,
          "kappa_positions", kappa_positions)

    print("sentence length:", statsstring(annotator_b["sentence_lengths"]))
    print("sentence length " + level + " annotator_a:",
          statsstring(annotator_a["sentence_level_lengths"]))
    print("sentence length " + level+  " annotator_b:",
          statsstring(annotator_b["sentence_level_lengths"]))


def main():
    if len(sys.argv) != 2:
        print("""Usage:
        scripts/compute_kappa.py [path to main directory, usually .]

        output:
        number of changes each annotator made,
        differences between the annotators,
        kappa (with string comparison),
        kappa (only checking whether a change was made to a token or not)
        """)
        exit(1)

    print("computing kappa for ZH1...")
    compute_kappa("ZH1")
    print("computing kappa for ZH2...")
    compute_kappa("ZH2")

    annotation_dirs = ["set1/xml", "set2/xml/annotator_a", "set2/xml/annotator_b"]

    print("\n########### overall ###########")
    print_stats_for_annotation_dirs(annotation_dirs, True)
    print("\n########### set 1 ###########")
    print_stats_for_annotation_dirs(["set1/xml"])
    print("\n########### set 2 annotator a ###########")
    print_stats_for_annotation_dirs(["set2/xml/annotator_a"])
    print("\n########### set 2 annotator b ###########")
    print_stats_for_annotation_dirs(["set2/xml/annotator_b"])

def print_stats_for_annotation_dirs(annotation_dirs, count_twice = False):
    for level in ["ZH1", "ZH2"]:
        leveldata = {}
        print("\ncomputing level " + level)
        if count_twice:
            print("Note that sentences from set 1 are counted twice to only half count separate annotations from set 2")
        for adir in annotation_dirs:
            print("  " + adir)
            dirdata = {}
            for fname in [sys.argv[1] + "/" + adir + "/" + x
                          for x in os.listdir(sys.argv[1] + "/" + adir)]:
                appendall(dirdata, read_xml_to_annotation(open(fname), level))
            appendall(leveldata, dirdata)
            if adir == "set1/xml" and count_twice:
                # append data twice to only half count separate annotations
                appendall(leveldata, dirdata)
            ctokstats = make_stats_from_list(dirdata["sentence_lengths"])
            lvlstats = make_stats_from_list(dirdata["sentence_level_lengths"])
            print("    ctok sentence length:",
                  stats_to_string(ctokstats))
            print("    " + level, "sentence length:",
                  stats_to_string(lvlstats))
        if count_twice:
            print("  " + "ctok sentence length:",
                  statsstring(leveldata["sentence_lengths"]))
            print("  " + level + " sentence length:",
                  statsstring(leveldata["sentence_level_lengths"]))


def statsstring(s):
    return stats_to_string(make_stats_from_list(s))


def make_stats_from_list(s):
    return {"mean": np.round(np.average(s), 1),
            "median": np.median(s),
            "25p": np.percentile(s, 25),
            "75p": np.percentile(s, 75),
            "sum": np.sum(s),
            "num": len(s)}


def stats_to_string(s):
    result = "mean: " + str(s["mean"])
    result += " median: " + str(s["median"])
    result += " 25% percentile: " + str(s["25p"])
    result += " 75% percentile: " + str(s["75p"])
    result += " sum (#tokens): " + str(s["sum"])
    result += " # elements: " + str(s["num"])
    return result


if __name__ == '__main__':
    main()
